import os

from redash.settings.helpers import parse_boolean, set_from_string


# Frequency of health query runs in minutes (12 hours by default)
HEALTH_QUERIES_REFRESH_SCHEDULE = int(
    os.environ.get("REDASH_HEALTH_QUERIES_REFRESH_SCHEDULE", 720)
)

# When enabled this will match the given remote groups request header with a
# configured list of allowed user groups.
REMOTE_GROUPS_ENABLED = parse_boolean(
    os.environ.get("REDASH_REMOTE_GROUPS_ENABLED", "false")
)
REMOTE_GROUPS_HEADER = os.environ.get(
    "REDASH_REMOTE_GROUPS_HEADER", "X-Forwarded-Remote-Groups"
)
REMOTE_GROUPS_ALLOWED = set_from_string(
    os.environ.get("REDASH_REMOTE_GROUPS_ALLOWED", "")
)



from pyhive import presto
from redash.query_runner import query_runners

from redash_stmo.query_runner.presto import STMOPresto, stmo_connect


def test_custom_presto_query_runner():
    assert query_runners['presto'] == STMOPresto


def test_custom_presto_connect():
    assert presto.connect == stmo_connect



from redash.handlers.query_results import QueryResultResource
from redash.permissions import require_access, require_permission, view_only
from redash.handlers.base import get_object_or_404
from redash import models

from ...resources import add_resource
from .parser import extract_table_names


class StmoQueryResultResource(QueryResultResource):
    @require_permission('view_query')
    def get(self, query_id=None, query_result_id=None, filetype='json'):
        if query_result_id:
            query_result = get_object_or_404(
                models.QueryResult.get_by_id_and_org,
                query_result_id,
                self.current_org,
            )
            # Look for queries matching this result whose data source is 'Query Results'.
            if models.Query.query.join(
                models.DataSource,
            ).filter(
                models.Query.query_hash == query_result.query_hash,
                models.DataSource.type == 'results',
            ).first():
                table_names = extract_table_names(query_result.query_text)
                for table_name in table_names:
                    # Look for query IDs being accessed.
                    if table_name.startswith("query_"):
                        try:
                            qid = int(table_name.split('_', 1)[1])
                        except ValueError:
                            # If it's not "query_NNN" it can't affect our permissions check here.
                            continue
                        upstream_q = models.Query.query.filter(
                            models.Query.id == qid
                        ).first()
                        if upstream_q is None:
                            continue
                        # If the user making this request doesn't have permission to
                        # view the query results being accessed in this query, deny
                        # access.
                        require_access(upstream_q.data_source.groups, self.current_user, view_only)

            require_access(query_result.data_source.groups, self.current_user, view_only)

        return super(StmoQueryResultResource, self).get(query_id, query_result_id, filetype)


def extension(app):
    # remove the original resource
    del app.view_functions['query_result']
    add_resource(
        app,
        StmoQueryResultResource,
        '/api/query_results/<query_result_id>.<filetype>',
        '/api/query_results/<query_result_id>',
        '/api/queries/<query_id>/results.<filetype>',
        '/api/queries/<query_id>/results/<query_result_id>.<filetype>',
        endpoint='query_result',
    )



from tests import BaseTestCase

from redash.models import db, Group


class TestQueryResultAPI(BaseTestCase):

    def test_query_result_query(self):
        restricted_group = self.factory.create_group(
            org=self.factory.org, type=Group.REGULAR_GROUP, name="restricted"
        )
        # Commit needed because User.group_ids is an array of IDs, not a relation
        db.session.commit()
        restricted_user = self.factory.create_user(
            group_ids=[restricted_group.id, self.factory.org.default_group.id],
            name="Restricted User",
        )
        restricted_ds = self.factory.create_data_source(
            group=restricted_group, view_only=False, type="pg"
        )
        upstream_query = self.factory.create_query(
            user=restricted_user,
            data_source=restricted_ds,
            query_text="select * from secret",
        )
        result_ds = self.factory.create_data_source(
            group=self.factory.org.default_group, view_only=False, type="results"
        )
        query = self.factory.create_query(
            user=restricted_user,
            data_source=result_ds,
            query_text="select password from query_{} limit 10".format(
                upstream_query.id
            ),
        )
        query_result = self.factory.create_query_result(
            data_source=result_ds,
            query_text=query.query_text,
            query_hash=query.query_hash,
        )
        rv = self.make_request(
            "get", "/api/query_results/{}".format(query_result.id), user=restricted_user
        )
        self.assertEquals(rv.status_code, 200)

    def test_query_result_query_restrict_access(self):
        restricted_group = self.factory.create_group(
            org=self.factory.org, type=Group.REGULAR_GROUP, name="restricted"
        )
        restricted_user = self.factory.create_user(
            group_ids=[restricted_group.id, self.factory.org.default_group.id]
        )
        restricted_ds = self.factory.create_data_source(
            group=restricted_group, view_only=False, type="pg"
        )
        upstream_query = self.factory.create_query(
            user=restricted_user,
            data_source=restricted_ds,
            query_text="select * from secret",
        )
        result_ds = self.factory.create_data_source(
            group=self.factory.org.default_group, view_only=False, type="results"
        )
        query = self.factory.create_query(
            user=restricted_user,
            data_source=result_ds,
            query_text="select password from query_{}".format(upstream_query.id),
        )
        query_result = self.factory.create_query_result(
            data_source=result_ds,
            query_text=query.query_text,
            query_hash=query.query_hash,
        )
        rv = self.make_request("get", "/api/query_results/{}".format(query_result.id))
        self.assertEquals(rv.status_code, 403)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2009-2018 the sqlparse authors and contributors
# <see AUTHORS file>
#
# This example is part of python-sqlparse and is released under
# the BSD License: https://opensource.org/licenses/BSD-3-Clause
#
# This example illustrates how to extract table names from nested
# SELECT statements.
#
# See:
# https://groups.google.com/forum/#!forum/sqlparse/browse_thread/thread/b0bd9a022e9d4895

import sqlparse
from sqlparse.sql import IdentifierList, Identifier
from sqlparse.tokens import Keyword, DML


def is_subselect(parsed):
    if not parsed.is_group:
        return False
    for item in parsed.tokens:
        if item.ttype is DML and item.value.upper() == 'SELECT':
            return True
    return False


def extract_from_part(parsed):
    from_seen = False
    for item in parsed.tokens:
        if from_seen:
            if is_subselect(item):
                for x in extract_from_part(item):
                    yield x
            elif item.ttype is Keyword:
                raise StopIteration
            else:
                yield item
        elif item.ttype is Keyword and item.value.upper() == 'FROM':
            from_seen = True


def extract_table_identifiers(token_stream):
    for item in token_stream:
        if isinstance(item, IdentifierList):
            for identifier in item.get_identifiers():
                yield identifier.get_real_name()
        elif isinstance(item, Identifier):
            yield item.get_real_name()
        # It's a bug to check for Keyword here, but in the example
        # above some tables names are identified as keywords...
        elif item.ttype is Keyword:
            yield item.value


def extract_table_names(sql):
    stream = extract_from_part(sqlparse.parse(sql)[0])
    return list(extract_table_identifiers(stream))



import subprocess

import apiclient.errors
from flask import request

from redash import models
from redash.handlers.base import BaseResource, get_object_or_404
from redash.permissions import not_view_only, has_access
from redash.utils import collect_parameters_from_request, json_loads, mustache_render

from redash_stmo.resources import add_resource


class DataSourceValidatorResource(BaseResource):

    def validate_pg(self, query_runner, query):
        p = subprocess.Popen("pgsanity", stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        stdout, _ = p.communicate(query + ";")
        return p.returncode == 0, stdout

    def validate_bigquery(self, query_runner, query):
        jobs = query_runner._get_bigquery_service().jobs()
        project_id = query_runner._get_project_id()
        job_data = {
            "configuration": {
                "query": {
                    "query": query,
                },
                "dryRun": True,
            }
        }

        if query_runner.configuration.get('useStandardSql', False):
            job_data['configuration']['query']['useLegacySql'] = False

        if query_runner.configuration.get('userDefinedFunctionResourceUri'):
            resource_uris = query_runner.configuration["userDefinedFunctionResourceUri"].split(',')
            job_data["configuration"]["query"]["userDefinedFunctionResources"] = map(
                lambda resource_uri: {"resourceUri": resource_uri}, resource_uris)

        if "maximumBillingTier" in query_runner.configuration:
            job_data["configuration"]["query"]["maximumBillingTier"] = query_runner.configuration["maximumBillingTier"]

        insert_response = jobs.insert(projectId=project_id, body=job_data).execute()
        if 'errorResult' in insert_response['status']:
            return False, insert_response['status']['errorResult']
        return True, "This query will process %s bytes." % (insert_response['statistics']['totalBytesProcessed'],)

    def get_validator(self, query_runner):
        """Return the query validator for the given query runner"""
        try:
            validator = getattr(self, 'validate_%s' % query_runner.type())
            if callable(validator):
                return validator
        except AttributeError:
            pass

    def post(self, data_source_id):
        params = request.get_json(force=True)
        parameter_values = collect_parameters_from_request(request.args)
        query = mustache_render(params['query'], parameter_values)
        data_source = get_object_or_404(
            models.DataSource.get_by_id_and_org,
            data_source_id,
            self.current_org,
        )

        # get the validator method
        validator = self.get_validator(data_source.query_runner)
        if (validator is None
                or not has_access(data_source.groups, self.current_user, not_view_only)):
            return {
                'valid': False,
                'report': 'You do not have permission to run queries with this data source.'
            }, 403

        try:
            valid, report = validator(data_source.query_runner, query)
            return {'valid': valid, 'report': report}, 200
        except apiclient.errors.HttpError as e:
            if e.resp.status == 400:
                error = json_loads(e.content)['error']['message']
            else:
                error = e.content
            return {'valid': False, 'report': error}, 200
        except Exception as e:
            return {'valid': False, 'report': str(e)}, 500


def extension(app):
    add_resource(app, DataSourceValidatorResource, '/api/data_sources/<data_source_id>/validate')



from flask import redirect, request, url_for
from redash import settings
from redash.authentication import get_next_path
from redash.authentication.org_resolving import current_org
from redash.authentication.remote_user_auth import logger

from redash_stmo import settings as extension_settings


def check_remote_groups():
    """Check if there is a header of user groups and if yes
    check it against a list of allowed user groups from the settings"""
    # Quick shortcut out if remote user auth or remote groups aren't enabled
    if (
        not settings.REMOTE_USER_LOGIN_ENABLED
        or not extension_settings.REMOTE_GROUPS_ENABLED
    ):
        return

    # Generate the URL to the remote auth login endpoint
    if settings.MULTI_ORG:
        org = current_org._get_current_object()
        remote_auth_path = url_for("remote_user_auth.login", org_slug=org.slug)
    else:
        remote_auth_path = url_for("remote_user_auth.login")

    # Then only act if the request is for the remote user auth view
    if request.path.startswith(remote_auth_path):
        remote_groups = settings.set_from_string(
            request.headers.get(extension_settings.REMOTE_GROUPS_HEADER) or ""
        )
        # Finally check if the remote groups found in the request header
        # intersect with the allowed remote groups
        if not extension_settings.REMOTE_GROUPS_ALLOWED.intersection(remote_groups):
            logger.error(
                "User groups provided in the %s header are not "
                "matching the allowed groups.",
                extension_settings.REMOTE_GROUPS_HEADER,
            )
            # Otherwise redirect back to the frontpage
            unsafe_next_path = request.args.get("next")
            next_path = get_next_path(unsafe_next_path)
            if settings.MULTI_ORG:
                org = current_org._get_current_object()
                index_url = url_for("redash.index", org_slug=org.slug, next=next_path)
            else:
                index_url = url_for("redash.index", next=next_path)
            return redirect(index_url)


def extension(app):
    """An extension that checks the REMOTE_GROUPS_HEADER."""
    app.before_request(check_remote_groups)



import collections
import logging

import six
from pyhive import presto
from redash.query_runner.presto import Presto
from redash.query_runner import register


logger = logging.getLogger(__name__)


class STMOPresto(Presto):
    """
    A custom Presto query runner. Currently empty.
    """


class STMOConnection(presto.Connection):
    """
    A custom Presto connection that uses the custom Presto cursor
    as the default cursor.
    """
    def cursor(self):
        return STMOPrestoCursor(*self._args, **self._kwargs)


class STMOPrestoCursor(presto.Cursor):
    """
    A custom Presto cursor that processes the data after it has been
    handled by the parent cursor to apply various transformations.
    """
    def _process_response(self, response):
        super(STMOPrestoCursor, self)._process_response(response)
        self._data = self._process_data()

    def _process_data(self):
        processed_data = collections.deque()
        for row in self._data:  # the top-level is an iterable of records (i.e. rows)
            item = []
            for column, row in zip(self._columns, row):
                item.append(self._format_data(column["typeSignature"], row))
            processed_data.append(item)
        return processed_data

    def _format_data(self, column, data):
        """Given a Presto column and its data, return a more human-readable
        format of its data for some data types."""
        type = column["rawType"]

        try:
            iter(data)  # check if the data is iterable
        except TypeError:
            return data  # non-iterables can simply be directly shown

        # records should have their fields associated with types
        # but keep the tuple format for backward-compatibility
        if type == "row":
            keys = column["literalArguments"]
            values = [
                self._format_data(c, d)
                for c, d in zip(column["typeArguments"], data)
            ]
            return tuple(zip(keys, values))

        # arrays should have their element types associated with each element
        elif type == "array":
            rep = [
                column["typeArguments"][0]
            ] * len(data)
            return (
                self._format_data(c, d)
                for c, d in zip(rep, data)
            )

        # maps should have their value types associated with each value
        # (note that keys are always strings), but keep the tuple format
        # for backward-compatibility
        elif type == "map":
            value_type = column["typeArguments"][1]
            return (
                (k, self._format_data(value_type, v))
                for k, v in six.iteritems(data)
            )

        else:
            # unknown type, don't process it
            return data


def stmo_connect(*args, **kwargs):
    """
    A custom connect function to be used to override the
    default pyhive.presto.connect function.
    """
    return STMOConnection(*args, **kwargs)


def extension(app):
    logger.info('Loading Redash Extension for the custom Presto query runner')
    # Monkeypatch the pyhive.presto.connect function
    presto.connect = stmo_connect
    # and register our own STMOPresto query runner class
    # which automatically overwrites the default presto query runner
    register(STMOPresto)
    logger.info('Loaded Redash Extension for the custom Presto query runner')
    return stmo_connect



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['redash_stmo',
 'redash_stmo.data_sources',
 'redash_stmo.data_sources.link',
 'redash_stmo.data_sources.validator',
 'redash_stmo.data_sources.version',
 'redash_stmo.handlers',
 'redash_stmo.handlers.authentication',
 'redash_stmo.handlers.query_results',
 'redash_stmo.query_runner']

package_data = \
{'': ['*'],
 'redash_stmo.data_sources.link': ['bundle/*'],
 'redash_stmo.data_sources.validator': ['bundle/*'],
 'redash_stmo.data_sources.version': ['bundle/*']}

install_requires = \
['dockerflow>=2018.4.0', 'requests', 'sqlparse', 'six', 'pyhive', 'pgsanity']

extras_require = \
{'dev': ['watchdog'], 'test': ['pytest', 'mock', 'pytest-cov', 'pytest-flake8']}

entry_points = \
{'redash.extensions': ['datasource_health = '
                       'redash_stmo.data_sources.health:extension',
                       'datasource_link = '
                       'redash_stmo.data_sources.link:extension',
                       'datasource_validator = '
                       'redash_stmo.data_sources.validator:extension',
                       'datasource_version = '
                       'redash_stmo.data_sources.version:extension',
                       'dockerflow = redash_stmo.dockerflow:extension',
                       'handler_queryresults = '
                       'redash_stmo.handlers.query_results:extension',
                       'handler_remote_user_auth = '
                       'redash_stmo.handlers.authentication.remote_user_auth:extension',
                       'queryrunner_presto = '
                       'redash_stmo.query_runner.presto:extension']}

setup(name='redash-stmo',
      version='2019.3.1',
      description='Extensions to Redash by Mozilla',
      author='Mozilla Foundation',
      author_email='dev-webdev@lists.mozilla.org',
      url='https://github.com/mozilla/redash-stmo',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
     )



import mock
from flask import url_for
from redash import settings

from redash_stmo import settings as extension_settings
from tests import BaseTestCase


class TestRemoteAuthGroups(BaseTestCase):
    @mock.patch.object(settings, "REMOTE_USER_LOGIN_ENABLED", return_value=True)
    @mock.patch.object(extension_settings, "REMOTE_GROUPS_ENABLED", return_value=True)
    @mock.patch.object(
        extension_settings,
        "REMOTE_GROUPS_ALLOWED",
        return_value=set(["admins", "managers"]),
    )
    @mock.patch("redash.authentication.remote_user_auth.logger.error")
    def test_redirect_if_disabled(self, mock_logger, *args, **kwargs):
        """Test to make sure requests to /login are directed to the
        remote auth URL"""
        next_path = "/"
        if settings.MULTI_ORG:
            test_url = url_for(
                "remote_user_auth.login", org_slug="default", next=next_path
            )
        else:
            test_url = url_for("remote_user_auth.login", next=next_path)

        with self.app.test_request_context(test_url):
            # make sure to call the before_request callback used
            self.app.preprocess_request()
            response = self.get_request(
                test_url, headers={"X-Forwarded-Remote-Groups": "staff,contributors"}
            )
            self.assertTrue(mock_logger.called)
            self.assertEqual(response.status_code, 302)
            index_url = url_for(
                "redash.index", org_slug="default", next=next_path, _external=True
            )
            self.assertEqual(response.location, index_url)
